/*
 * Pam routine to call your favorite routine 
 * to get an AFS token. This could be aklog, afslogin, or gssklog
 * The routine can then use the environment to find
 * the GSI or KRB5 environment variables. 
 * 
 * If a previous PAM routine, like pam_krb5 or the sshd
 * has set the name in the pam_env, then we will pass that on
 * via the env. 
 * 
 * Since hpux does not have the pam_env functions, we will
 * assume the previous ssh or pam_krb5 has set the name
 * of the ccache using either the default of /tmp/krb5cc_<uid>
 * or followed some pattern.  We can pass in the pattern
 * to both pam exists. via ccache=<name>, for example
 *  ccache=FILE:/tmp/krb5cc_p%p_u%u 
 * where %p  and %u are replaced with the PID and UID.
 *
 * Since many applications and systems dont really use PAM
 * correctly, we can get a PAG and token in 
 * pam_sm_authenticate, pam_sm_setcreds, or pam_sm_open_session.
 * Parameters that can be used:
 *  nopag - no pag will be obtained
 *  force_afs - get pag and token during the authenticate
 *              rather then setcreds 
 * 
 * So only add the session or the authenticate entries to
 * to the pam.conf don't add both for the same application. 

 * If needed the pam_krb5 authenticate can be forced to store
 * the creds early. 
 */

#define DEBLOG(A,B) \
	if (debug) \
		syslog(LOG_DEBUG,"pam_afs2:%d %s:%s", __LINE__, A, B)

		
#include <stdio.h>
#include <pam_appl.h>
#include <pam_modules.h>
#include <syslog.h>
#include <pwd.h>

int 
do_afs_routine(pam_handle_t *pamh, int flags, int argc,
	const char **argv);


/***************************************************************/

int
pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc,
	    const char **argv)
{

	int debug = 0;
	int i;
	int ret;
	int do_afs= 0;
	
	for (i = 0; i < argc; i++) {
		if (strcmp(argv[i], "debug") == 0) {
			debug++;
		}
		else if (strcmp(argv[i], "force_afs") == 0) {
			do_afs = 1;
		}
	}
#ifdef DEBUG
	fprintf(stderr,"pam_sm_authenticate flag=%d\n",flags);
#endif
	DEBLOG("pam_sm_authenticate","called");

	if (do_afs) {
		ret = do_afs_routine(pamh, flags, argc, argv);

		if (ret == PAM_SUCCESS) {
			ret = PAM_IGNORE; /* we did not really authenticate */
		}
	} else {
		ret = PAM_IGNORE;
	}

	return ret;
}

/********************************************************/ 

int 
pam_sm_setcred(pam_handle_t *pamh, int flags, int argc,
           const char **argv)
{
    int debug = 0;
    int ret;
    int i;
	int do_afs = 1;  /* default to doing the afs stuff */

    for (i = 0; i < argc; i++) {
        if (strcmp(argv[i], "debug") == 0) {
            debug++;
        }
		else if (strcmp(argv[i], "force_afs" ) == 0) {
			do_afs = 0; /* authenticate would have done this */
		}
			
    }

    DEBLOG("pam_sm_setcred", "called");
	
	if (do_afs) {
		ret = do_afs_routine(pamh, flags, argc, argv);
	} else {
		ret = PAM_SUCCESS;
	}
	return ret;
}

/********************************************************/ 

int 
pam_sm_open_session(pam_handle_t *pamh, int flags, int argc,
           const char **argv)
{
    int debug = 0;
	int i;
	int ret;
	

#ifdef DEBUG
    fprintf(stderr,"pam_sm_open_session flag=%d\n",flags);
#endif

    for (i = 0; i < argc; i++) {
        if (strcmp(argv[i], "debug") == 0) {
            debug++;
        }
	}

    DEBLOG("pam_sm_open_session", "called");

	 ret = do_afs_routine(pamh, flags, argc, argv);

    return ret;
}

/********************************************************/


int
pam_sm_close_session(pam_handle_t *pamh, int flags, int argc,
           const char **argv)
{
    int debug = 0;
    int ret;
    int i;

    for (i = 0; i < argc; i++) {
        if (strcmp(argv[i], "debug") == 0) {
            debug++;
        }
    }

    DEBLOG("pam_sm_close_session", "called");

    return PAM_SUCCESS;
}



/********************************************************/ 

int 
do_afs_routine(pam_handle_t *pamh, int flags, int argc,
           const char **argv)
{

	int debug = 0;
	int set_pag = 1;
	int i;
	int ret;
	char * username = NULL;
    char ** env;
	char * pgm_name = NULL;
	struct passwd   *pw = NULL;

#ifdef __hpux
	char *p;
	char *q;
	char *cache_name = NULL;
	char cache_name_buf[1024];
	char cache_name_env_buf[2048];
    char *hpenv_path = "PATH=/krb5/sbin:/usr/bin:/bin";
    char *hpenv[] = {0,0,0};
#endif
	
#ifdef DEBUG
	fprintf(stderr,"do_afs_routine flag=%d\n",flags);
#endif

	for (i = 0; i < argc; i++) {
		if (strcmp(argv[i], "debug") == 0) {
			debug++;
		}
#ifdef __hpux
		else if (strncmp(argv[i], "ccache=", 7) == 0) {
			cache_name = &argv[i][7]; /* save for later */
		}
#endif
        else if (strncmp(argv[i], "pgm=", 4) == 0) {
            pgm_name = &argv[i][4]; /* save for later */
        }

#ifdef __aix
		else if (strcmp(argv[i], "-a") == 0) {
			apagaix_exec = argv[i+1];
			i++;
		}
#endif
		else if (strcmp(argv[i], "nopag")  == 0) {
			set_pag = 0;
		}
	}

	DEBLOG("do_afs_routine", "called");
	
#if 0
	if (!(flags & PAM_ESTABLISH_CRED)) {
		ret = PAM_SUCCESS;
		DEBLOG("do_afs_routine", "no  PAM_ESTABLISH_CRED");
		goto err;
	}
#endif

	if (pam_get_item(pamh, PAM_USER, (void **) &username)) {
		ret = PAM_SERVICE_ERR;
		goto err;
	}

	if (strncmp(username,"root",4) == 0) {
		ret = PAM_IGNORE; /* no tokens for root */
		DEBLOG("do_afs_routine ignoring ",username);
		goto err;
	}

	DEBLOG("do_afs_routine user", username);

    if (!(pw = getpwnam(username))) {
        ret = PAM_IGNORE;
		DEBLOG("do_afs_routine" , "user not known");
        goto err;
    }

#ifdef __hpux

	/* HP does not have the pam_env routines */
	/* will look for ccache as set in krb5 pam routines */

	/* Get the cache name */
	if (!cache_name) {
		cache_name = "FILE:/tmp/krb5cc_%u";
	}
	strcpy(cache_name_env_buf,"KRB5CCNAME=");
	p = cache_name_env_buf + strlen(cache_name_env_buf);
	q = cache_name;

	/* convert %u and %p */
	while (*q) {
		if (*q == '%') {
			q++;
			if (*q == 'u') {
				sprintf(p, "%ld", pw->pw_uid);
				p += strlen(p);
			} else if (*q == 'p') {
				sprintf(p, "%ld", getpid());
				p += strlen(p);
			} else {
				/* Not a special token */
				*p++ = '%';
				q--;
			}
			q++;
		} else {
			*p++ = *q++;
		}
	}

	hpenv[0] = hpenv_path;
	hpenv[1] = cache_name_env_buf;
	env = hpenv;
#else
	env = pam_getenvlist(pamh);
#endif

#ifdef DEBUG
    fprintf(stderr,"do_afs_routine pid=%d uid=%d euid=%d\n",
        getpid(),getuid(),geteuid());
     fprintf(stderr,"do_afs_routine, pw_dir=%s\n",
        pw->pw_dir);
#ifndef __hpux
     fprintf(stderr,"do_afs_routine Kenv=%s\n",
        pam_getenv(pamh,"KRB5CCNAME")?
        pam_getenv(pamh,"KRB5CCNAME"):"(none)");
    if (debug) {
		if (pam_getenv(pamh,"KRB5CCNAME")) 
			DEBLOG("pam_getenv KRB5CCNAME", 
					pam_getenv(pamh,"KRB5CCNAME"));
		if (getenv("KRB5CCNAME")) 
			DEBLOG("getenv KRB5CCNAME",
					getenv("KRB5CCNAME"));
	}
#endif

#endif

	DEBLOG("do_afs_routine","calling get_afs_token");
    get_afs_token(pgm_name, env, pw->pw_uid, pw->pw_dir, set_pag);
	
	/* if we get this far return success, even if failed  to get token */
	ret = PAM_SUCCESS;

err:
	return ret;
}
